
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#ifndef PG_HTTP_HEADERS_H
#define PG_HTTP_HEADERS_H

#include "xstd/h/iostream.h"

#include "xstd/Array.h"
#include "xstd/NetAddr.h"
#include "xstd/Checksum.h"
#include "base/ObjId.h"
#include "runtime/ObjWorld.h"
#include "runtime/XactAbortCoord.h"


class MsgHdrParsTab;

class HttpUri {
	public:
		HttpUri(): pathBuf(0), pathLen(-1) {}

	public:
		ObjId oid;
		NetAddr host;
		const char *pathBuf;
		int pathLen;
};

// common interface for parsing HTTP requests and responses
class MsgHdr {
	public:
		typedef bool (MsgHdr::*Parser)(const char *buf, const char *eoh);

	public:
		static bool ParseNetAddr(const char *buf, const char *eoh, NetAddr &addr);
		static bool ParseDname(const char *buf, const char *eoh, NetAddr &addr);
		static bool ParseIp(const char *buf, const char *eoh, NetAddr &addr);
		static bool ParseHostInUri(const char *&start, const char *eorl, NetAddr &host);

	public:
		MsgHdr(const MsgHdrParsTab &aTab);
		virtual ~MsgHdr() {}

		virtual void reset();

		bool keepAlive() const;

		bool parse(const char *buf, Size sz);
		bool parseUri(const char *&buf, const char *eoh, HttpUri &uri);

		virtual void store(OLog &log) const;

	protected:
		static void Configure(MsgHdrParsTab &tab);
		static int AddParser(const String &field, Parser parser, MsgHdrParsTab &tab);

	protected:
		void parseFields();
		bool parseUniqId(const char *&buf, const char *eoh, UniqId &id);
		bool parseWorld(const char *buf, const char *eoh, ObjWorld &world);

		virtual bool parseRLine(const char *buf, const char *eol) = 0;

		// can be used by both requests and replies
		bool parseDate(const char *buf, const char *eoh);
		bool parseContLen(const char *buf, const char *eoh);
		bool parseContMd5(const char *buf, const char *eoh);
		bool parsePragma(const char *buf, const char *eoh);
		bool parseCControl(const char *buf, const char *eoh);
		bool parseXXact(const char *buf, const char *eoh);
		bool parseXTarget(const char *buf, const char *eoh);
		bool parseXRemWorld(const char *buf, const char *eoh);
		bool parseXAbort(const char *buf, const char *eoh);
		bool parseXPhaseSyncPos(const char *buf, const char *eoh);
		bool parseConnection(const char *buf, const char *eoh);

		// these have to be here so we can register them in req
		virtual bool parseGetReqLine(const char *buf, const char *eorl);
		virtual bool parseHeadReqLine(const char *buf, const char *eorl);
		virtual bool parsePostReqLine(const char *buf, const char *eorl);
		virtual bool parsePutReqLine(const char *buf, const char *eorl);
		virtual bool parseHost(const char *buf, const char *eoh);
		virtual bool parseIms(const char *buf, const char *eoh);
		virtual bool parseXLocWorld(const char *buf, const char *eoh);
		

		// these have to be here so we can register them in rep
		virtual bool parseLocation(const char *buf, const char *eoh);
		virtual bool parseLMT(const char *buf, const char *eoh);
		virtual bool parseExpires(const char *buf, const char *eoh);

	public:
		const MsgHdrParsTab &theParsTab;

		Size theHdrSize;
		Time theDate;          // Date: header field
		Size theContSize;
		Checksum theChecksum;  // Content-MD5

		UniqId theGroupId;
		UniqId theXactId;
		NetAddr theTarget;

		ObjWorld theRemWorld; // remote object world specs
		XactAbortCoord theAbortCoord; // abort coordinates
		int thePhaseSyncPos; // index of a ready-to-stop phase

		double theHttpVer;
		enum { kaNo, kaDefault, kaYes } theKeepAliveConn;
		bool isCachable;

	protected:
		Array<const char *> theFields; // positions of header fields
		const char *theBufBeg, *theBufEnd;
		const char *theSrchPtr;        // parse buffer iterator
		enum { ssFirst, ssSkip, ssFound } theSrchState;
};

class ReqHdr: public MsgHdr {
	public:
		static void Configure();
		static void Clean();
		
	public:
		ReqHdr();

		virtual void reset();

		virtual void store(OLog &log) const;

	protected:
		virtual bool parseRLine(const char *buf, const char *eol);

		bool parseAnyReqLine(const char *buf, const char *eorl);
		virtual bool parseGetReqLine(const char *buf, const char *eorl);
		virtual bool parseHeadReqLine(const char *buf, const char *eorl);
		virtual bool parsePostReqLine(const char *buf, const char *eorl);
		virtual bool parsePutReqLine(const char *buf, const char *eorl);
		virtual bool parseHost(const char *buf, const char *eoh);
		virtual bool parseIms(const char *buf, const char *eoh);
		virtual bool parseXLocWorld(const char *buf, const char *eoh);

	protected:
		static MsgHdrParsTab *TheParsTab;

	public:
		HttpUri theUri;
		Time theIms;          // If-Modified-Since timestamp
		ObjWorld theLocWorld; // local world info
		bool isHealthCheck;
		bool expectBody() const;
};

class RepHdr: public MsgHdr {
	public:
		enum StatusCode { scUnknown = -1, sc200_OK = 200, sc204_NoContent = 204,
			sc300_Choices = 300, sc302_Found = 302, sc303_Other = 303,
			sc304_NotModified = 304, sc307_Tmp_Redir = 307};

	public:
		static void Configure();
		static void Clean();
		
	public:
		RepHdr();

		virtual void reset();

		bool polyHeaders() const { return theXactId; }
		bool expectPolyHeaders() const { return expectBody() && theStatus != 302; }
		bool expectBody() const;
		Time calcLmt() const;
		bool redirect() const;

		virtual void store(OLog &log) const;

	protected:
		virtual bool parseRLine(const char *buf, const char *eol);

		virtual bool parseLocation(const char *buf, const char *eoh);
		virtual bool parseLMT(const char *buf, const char *eoh);
		virtual bool parseExpires(const char *buf, const char *eoh);

	protected:
		static MsgHdrParsTab *TheParsTab;

	public:
		HttpUri theLocn;  // Location: in 3xx responses
		Time theLMT;      // Last-Modified-Time header field
		Time theExpires;  // Expires HTTP header field
		int theStatus;    // HTTP status code
};

inline ostream &crlf(ostream &os) { return os.write("\r\n", 2); }

#endif
