
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#include "base/polygraph.h"

#include <limits.h>
#include "xstd/h/strstream.h"

#include "xstd/Clock.h"
#include "xstd/Rnd.h"
#include "xstd/Checksum.h"
#include "xstd/gadgets.h"
#include "base/RndPermut.h"
#include "base/ObjId.h"
#include "base/polyLogCats.h"
#include "runtime/Connection.h"
#include "runtime/AddrMap.h"
#include "runtime/HostMap.h"
#include "runtime/Viserv.h"
#include "runtime/PubWorld.h"
#include "runtime/HttpDate.h"
#include "runtime/ErrorMgr.h"
#include "runtime/SharedOpts.h"
#include "runtime/PopModel.h"
#include "runtime/StatPhaseSync.h"
#include "runtime/LogComment.h"
#include "runtime/httpText.h"
#include "runtime/polyBcastChannels.h"
#include "runtime/polyErrors.h"
#include "runtime/globals.h"
#include "csm/ContentMgr.h"
#include "csm/ContentCfg.h"
#include "csm/BodyIter.h"
#include "csm/ObjLifeCycle.h"
#include "csm/oid2Url.h"
#include "server/Server.h"
#include "server/SrvCfg.h"
#include "server/SrvXact.h"
#include "server/SrvOpts.h"


void SrvXact::reset() {
	Xaction::reset();
	theOwner = 0;
	theContentCfg = 0;
	theBodyIter = 0;
	theReqHdr.reset();
	theTimes.reset();
	theRepHdrSize = -1;
	theProducedSize = 0;
	theConsumedSize = 0;

	theLogCat = lgcSrvSide;
}

Agent *SrvXact::owner() {
	return theOwner;
}

void SrvXact::exec(Server *anOwner, Connection *aConn, Time delay) {
	Assert(anOwner);
	theOwner = anOwner;
	theConn = aConn;

	if (delay > 0)
		sleepFor(delay);
	else
		doStart();
}

void SrvXact::doStart() {
	start(theConn);

	newState(stHdrWaiting);

	Assert(!theConn->theRd.theReserv); // cleared in conn mgr
	// note: we assume that conn mgr read() on first noteReadReady()
	noteDataReady(); // push state machine forward
}

void SrvXact::finish(Error err) {
	Xaction::finish(err);
	if (theBodyIter)
		theContentCfg->putBodyIter(theBodyIter);
	theOwner->noteXactDone(this);
}

void SrvXact::wakeUp(const Alarm &a) {
	AlarmUser::wakeUp(a);
	doStart();
}

void SrvXact::noteReadReady(int) {
	if (abortIo(&Connection::read))
		return;

	if (theConn->bad())
		finish(errOther);
	else
		noteDataReady();
}

void SrvXact::noteDataReady() {
	if (theState == stHdrWaiting) {
		if (theReqHdr.parse(theConn->theRdBuf.content(), theConn->theRdBuf.contSize())) {
			if (const Error err = interpretHeader()) {
				finish(err);
				return;
			}
			newState(stBodyWaiting); // if any
		} else
		if (theConn->theRdBuf.full()) { // header too big
			finish(errHugeHdr);
			return;
		} else
		if (theConn->atEof()) {         // premature end of headers
			finish(errPrematureEoh);
			return;
		}
	}

	if (theState == stBodyWaiting) {
		consume(theConn->theRdBuf.contSize());
		theConn->theRdBuf.pack();
		if (theConsumedSize == theReqSize) { // got everything expected
			theConn->theRd.stop(this);
			newState(stSpaceWaiting);
		} else
		if (theConn->atEof()) { // premature eof
			if (cfgAbortedReq()) {
				theOid.aborted(true);
				finish(0); // not an error, configuration told client to abort
				return;
			} else {
				finish(errPrematureEof);
				return;
			}
		}
	}

	if (theState == stHdrWaiting || theState == stBodyWaiting) {
		if (!theConn->theRd.theReserv)  // need to read more
			theConn->theRd.start(this);
	} else
	if (theState == stSpaceWaiting) {
		if (!theConn->theWr.theReserv)  // need to write more
			theConn->theWr.start(this);
	}
}

void SrvXact::consume(Size size) {
	theConn->theRdBuf.consumed(size);
	theConsumedSize += size;
}

Error SrvXact::interpretHeader() {
	// XXX: pxy server cannot support ignoreUrls; see Server::hostIdx
	const bool ignoreUrls = theReqHdr.isHealthCheck ||
		(TheSrvOpts.ignoreUrls && theOwner->hostIdx() >= 0);

	if (ignoreUrls) {
		// XXX: theOid.visName() will not be set!
		// we have to ignore host settings;
		// should we create our own worlds??
		static RndGen rng;
		theReqHdr.theUri.oid.world().create();
		theReqHdr.theUri.oid.target(theOwner->hostIdx());
		theReqHdr.theUri.oid.type(1);
		theReqHdr.theUri.oid.name(rng.ltrial() | 1);
		theReqHdr.theUri.oid.foreignSrc(false);
		theReqHdr.theUri.oid.foreignUrl(0);

		// XXX: report healthchecks once in a while and collect stats
		if (theReqHdr.isHealthCheck) {
			static bool didOnce = false;
			if (!didOnce) {
				Comment(6) << "fyi: first health check received from " << theConn->raddr() << ':' << endc;
				printMsg(theConn->theRdBuf, theReqHdr.theHdrSize);
				didOnce = true;
			}
		}
	}

	// dump request header
	if (TheOpts.theDumpFlags(dumpReq, dumpHdr))
		printMsg(theConn->theRdBuf, theReqHdr.theHdrSize);

	theOid = theReqHdr.theUri.oid;

	if (theOid.foreignUrl())
		return errForeignUrl;

	// calculate request size
	if (theReqHdr.expectBody()) {
		if (theReqHdr.theContSize < 0)
			return errNoCLen;
		theReqSize = theReqHdr.theHdrSize + theReqHdr.theContSize;
	} else {
		if (theReqHdr.theContSize >= 0)
			return errUnexpectedCLen;
		theReqSize = theReqHdr.theHdrSize;
	}

	if (!ignoreUrls) {
		if (!theReqHdr.theUri.host)
			return errNoHostName;

		if (const Error err = setViserv(theReqHdr.theUri.host))
			return err; // setViserv called finish

		if (!theReqHdr.theTarget)
			return errNoTarget;

		if (const Error err = setTarget(theReqHdr.theTarget))
			return err; // setTarget called finish
		Assert(theOid.target() >= 0);

		if (const Error err = checkUri()) // must be done before we consume()
			return err; // checkUri called finish
	}

	// note: clt and srv code assume we consume only after parse!
	consume(theReqHdr.theHdrSize);
	theConn->theRdBuf.pack();

	// update history for this server
	if (theOid.viserv() >= 0 && theReqHdr.theLocWorld)
		updatePubWorld(theReqHdr.theLocWorld);

	// update client group info
	if (theReqHdr.thePhaseSyncPos >= 0 && theReqHdr.theGroupId)
		TheStatPhaseSync.notePhaseSync(theReqHdr.theGroupId, theReqHdr.thePhaseSyncPos);

	// note if the client will close the connection
	theConn->lastUse(!theReqHdr.keepAlive());

	theOid.reload(!theReqHdr.isCachable);

	// get content specs for this reply
	Assert(theOid.type() > 0);
	theContentCfg = TheContentMgr.get(theOid.type());

	theRepHdrSize = -1; // so that we know that we need to make a reply

	return 0;
}

void SrvXact::noteWriteReady(int) {
	Assert(theState == stSpaceWaiting);
	WrBuf &buf = theConn->theWrBuf;

	if (theRepHdrSize < 0)
		makeRep(buf);

	Assert(theProducedSize > 0 || buf.contSize() > 0);
	
	// keep the write buffer full
	if (theBodyIter)
		theBodyIter->pour();
	else
		Assert(theProducedSize + buf.contSize() >= theRepSize);

	Size sz;
	if (abortIo(&Connection::write, &sz))
		return;

	if (theConn->bad()) {
		finish(errOther);
		return;
	}

	Assert(sz >= 0);
	theProducedSize += sz;
	if (theProducedSize >= theRepSize) { // done
		Assert(theProducedSize == theRepSize);
		finish(0);
		return;
	}

	if (!theConn->theWr.theReserv) // need write more
		theConn->theWr.start(this);
}

// decide what kind of reply to build
void SrvXact::makeRep(WrBuf &buf) {
	Assert(theContentCfg);

	// timestamps, cachability, aborts may be used for all types of replies
	theContentCfg->calcTimes(theOid, theTimes);
	theOid.cachable(theContentCfg->calcCachability(theOid));
	theOwner->cfg()->selectAbortCoord(theAbortCoord);

	theOwner->selectRepType(theOid);

	const char *repStart = buf.space();
	ostrstream os(buf.space(), buf.spaceSize());

	if (shouldMake302Found() && make302Found(os))
		; // nothing to be done here
	else
	if (shouldMake304NotMod())
		make304NotMod(os);
	else
		make200OK(os);

	buf.appended(os.tellp());

	Assert(theRepHdrSize > 0 && theRepSize >= theRepHdrSize);

	theAbortSize = theAbortCoord.pos(theRepHdrSize, theRepSize-theRepHdrSize);

	// dump reply header
	if (TheOpts.theDumpFlags(dumpRep, dumpHdr))
		printMsg(repStart, theRepHdrSize);
}

bool SrvXact::shouldMake302Found() const {
	return theOid.repToRedir();
}

bool SrvXact::canMake302Found(ObjId &oid) const {
	Assert(theOwner->popModel());

	// at this time redirect to the same server only
	if (theOid.viserv() < 0)
		return false;

	// misconfiguration
	if (!theOwner->popModel())
		return false;

	PubWorld &world = *ThePubWorlds[theOid.viserv()];

	if (!world.canRepeat())
		return false;

	oid = theOid; // set viserv and such
	oid.name(-1);
	oid.type(-1); // overwrite name and type
	world.repeat(oid, theOwner->popModel());
	oid.type(Oid2ContType(theOid));
	return true;
}

bool SrvXact::make302Found(ostream &os) {
	ObjId newOid;

	// make sure we have an object to redirect to
	if (!canMake302Found(newOid)) {
		ReportError(errMake302Found);
		return false;
	}

	theHttpStatus = RepHdr::sc302_Found;

	os << slHttp_10_302;
	putStdFields(os);

	os << hfpLocation;
	const Size urlStart = os.tellp();
	Oid2Url(newOid, os);
	const Size urlLen = Size(os.tellp()) - urlStart;
	os << crlf;

	const Size clen = urlLen + Size(text302Found.len());
	os << hfpContLength << (int)clen << crlf;

	putXFields(os);

	os << crlf; // end-of-headers

	// we determine the type of IMS request later
	theRepHdrSize = os.tellp();

	// text for humans
	Oid2Url(newOid, os << text302Found);

	theRepSize = theRepHdrSize + clen;
	return true;
}

bool SrvXact::shouldMake304NotMod() const {
	return 
		theReqHdr.theIms >= 0 && theTimes.lmt() >= 0 && 
		theTimes.lmt() <= theReqHdr.theIms;
}

void SrvXact::make304NotMod(ostream &os) {
	theHttpStatus = RepHdr::sc304_NotModified;

	os << slHttp_10_304;
	putStdFields(os);
	os << crlf; // end-of-headers

	theRepHdrSize = theRepSize = os.tellp();
	theOid.ims304(true);
}

// build headers for a happy "200 OK" reply
void SrvXact::make200OK(ostream &os) {
	theHttpStatus = RepHdr::sc200_OK;

	os << slHttp_10_200;
	const Size clen = put200OkHead(os);
	os << crlf; // end-of-headers

	theRepHdrSize = os.tellp();
	if (theOid.head()) {
		// do not send message body for HEAD
		theRepSize = theRepHdrSize;
	} else {
		theBodyIter = theContentCfg->getBodyIter(theOid);
		theBodyIter->configure(&theConn->theWrBuf, clen);
		theRepSize = theRepHdrSize + clen;
	}

	// note: non-200 and non-304 responses to IMS requests not covered yet
	theOid.ims200(theReqHdr.theIms >= 0);
}

Size SrvXact::put200OkHead(ostream &os) {
	/* it is "good practice" to send general-header fields first,
	 * followed by request-header or response-header fields, and
	 * ending with the entity-header fields. */

	/* general-header fields */

	os << (theOid.cachable() ? hfCcCachable : hfCcUncachable);

	putStdFields(os);

	/* other entity-header fields */

	if (theOid.cachable() && theTimes.showLmt()) 
		HttpDatePrint(os << hfpLmt, theTimes.lmt()) << crlf;

	const Size clen = theContentCfg->calcRepSize(theOid);
	os << hfpContLength << (int)clen << crlf;

	if (theContentCfg->theMimeType)
		os << hfpContType << theContentCfg->theMimeType << crlf;

	if (theContentCfg->calcChecksumNeed(theOid))
		putChecksum(os, clen);

	/* extention-header fields (they are entity-header fields as well) */
	putXFields(os);

	return clen;
}

// put well-known fields acceptible for both 304 and 200 replies
void SrvXact::putStdFields(ostream &os) const {
	// general-header fields

	HttpDatePrint(os << hfpDate) << crlf;

	// indicate persistency (HTTP/1.0 defaults to non-persistent)
	if (theConn->reusable())
		os << hfConnAliveOrg;

	// response-header fields (none)
	// entity-header fields

	if (theOid.cachable() && theTimes.knownExp()) 
		HttpDatePrint(os << hfpExpires, theTimes.exp()) << crlf;
}

// put extension header fields acceptible for both 302 and 200 replies
void SrvXact::putXFields(ostream &os) const {
	// put group ids with source/target ids on one line?
	os << hfpXTarget << theOwner->host() << crlf;

	if (theReqHdr.theXactId)
		os << hfpXXact 
			<< TheGroupId
			<< ' ' << theReqHdr.theXactId.genMutant()
			<< crlf;

	if (theOid.viserv() >= 0 && theReqHdr.theRemWorld)
		putRemWorld(os, theReqHdr.theRemWorld);

	os << hfpXAbort	<< ' ' << theAbortCoord.whether()
		<< ' ' << theAbortCoord.where()	<< crlf;

	os << hfpXPhaseSyncPos << TheStatPhaseSync.phaseSyncPos() << crlf;
}

// build and put checksum header field
void SrvXact::putChecksum(ostream &os, Size clen) const {
	ChecksumAlg a;

	BodyIter *i = theContentCfg->getBodyIter(theOid);
	WrBuf buf;
	i->configure(&buf, clen);
	while (*i) {
		i->pour();
		a.update(buf.content(), buf.contSize());
		buf.reset();
	}
	a.final();
	theContentCfg->putBodyIter(i);

	os << hfpContMd5;
	PrintBase64(os, a.sum().image(), a.sum().size());
	os << crlf;
}

Error SrvXact::setViserv(const NetAddr &name) {
	// XXX: the lookups below are very slow
	int nameIdx = -1;
	if (!TheAddrMap->find(name, nameIdx))
		return errForeignHostName;

	int viserv = -1;
	if (!TheViservs.find(nameIdx, viserv)) {
		TheViservs.add(nameIdx);
		ThePubWorlds.append(new PubWorld());
	}

	theOid.viserv(viserv);
	return 0;
}

Error SrvXact::setTarget(const NetAddr &) {
	// XXX: the check below should know about DNS RR, etc.
	//if (target != theOwner->host())
	//	ReportError(errMisdirRequest);

	theOid.target(theOwner->hostIdx());
	return 0;
}

// merge with CltXact::cfgAbortedReply()?
bool SrvXact::cfgAbortedReq() const {
	if (!theReqHdr.theAbortCoord)
		return false;

	const Size abSz = theReqHdr.theAbortCoord.pos(theReqHdr.theHdrSize, theReqSize-theReqHdr.theHdrSize);
	if (!Should(abSz >= 0))
		return false;

	if (abSz > theConsumedSize)
		return false;

	return true;
}


// check that requested URI belongs to our server
Error SrvXact::checkUri() {
	const int pathLen = theReqHdr.theUri.pathLen;
	Assert(pathLen >= 0);
	Assert(pathLen <= theConn->theRdBuf.contSize());
	Assert(theReqHdr.theUri.pathBuf);

	// first check if the content type belongs to us
	if (!theOwner->cfg()->hasContType(theOid.type())) {
		if (ReportError(errMisdirRequest)) {
			Comment << "host " << theOwner->host()
				<< " does not have content type implied by request URI: ";
			Comment.write(theReqHdr.theUri.pathBuf, pathLen);
			Comment << endc;
		}
		return errOther;
	}

	ostrstream os; // XXX: expensive mallocs
	Oid2UrlPath(theOid, os);
	os << ends;

	const char *buf = os.str();
	const bool res = (pathLen+1 == (int)os.tellp()) &&
		strncmp(buf, theReqHdr.theUri.pathBuf, pathLen) == 0;
	ostrstreamFreeze(os, false);

	if (!res) {
		if (ReportError(errMisdirRequest)) {
			Comment << "host: " << theOwner->host() << endc;
			(Comment << "received: ").write(theReqHdr.theUri.pathBuf, pathLen);
			Comment << endc;
			Oid2UrlPath(theOid, Comment << "expected: ");
			Comment << endc;
		}
		return errOther;
	}
	return 0;
}

void SrvXact::updatePubWorld(const ObjWorld &newSlice) {
	PubWorld &world = *ThePubWorlds[theOid.viserv()];

	int sliceIdx;
	if (!world.find(newSlice.id(), sliceIdx)) {
		world.add(newSlice);
	} else {
		world.sliceAt(sliceIdx).update(newSlice);
	}
}

void SrvXact::putRemWorld(ostream &os, const ObjWorld &oldSlice) const {
	PubWorld &world = *ThePubWorlds[theOid.viserv()];

	int sliceIdx;
	if (world.find(oldSlice.id(), sliceIdx)) {
		if (const PubWorldSlice *slice = world.newerSlice(oldSlice, sliceIdx))
			os << hfpXRemWorld << *slice << crlf;
	} 

	// else client-side sent remote world ID before the same public world ID
}

void SrvXact::logStats(OLog &ol) const {
	Xaction::logStats(ol);
	ol << theOwner->seqvId();
	ol << (int)theReqHdr.theIms.sec();
}
