
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#include "base/polygraph.h"

#include "xstd/h/iostream.h"
#include "xstd/h/strstream.h"
#include <fstream.h>
#include "xstd/h/iomanip.h"

#include "xstd/gadgets.h"
#include "base/CmdLine.h"
#include "base/opts.h"
#include "base/polyOpts.h"
#include "pgl/PglArraySym.h"
#include "pgl/PglNetAddrSym.h"
#include "pgl/PglNetAddrRange.h"
#include "pgl/PglNetAddrRangeSym.h"
#include "pgl/AddrMapSym.h"
#include "pgl/PglPp.h"
#include "pgl/PglParser.h"
#include "pgl/PglSemx.h"


class MyOpts: public OptGrp {
	public:
		MyOpts():
			theHelpOpt(this,    "help",          "list of options"),
			theVersOpt(this,    "version",       "package version info"),
			theCfgName(this,    "config <filename>",  "PGL configuration"),
			theCfgDirs(this,    "cfg_dirs <dirs>",  "directories for PGL #includes"),
			theZoneName(this,   "zone <str>",    "zone name"),
			theAddrStr(this,    "addresses <str>","IP addresses"),
			theNameStr(this,    "names <str>",   "domain names")
			{}

		virtual bool validate() const;

		//virtual ostream &printAnonym(ostream &os) const;
		//virtual bool parseAnonym(const Array<const char *> &opts);
		//virtual bool canParseAnonym() const { return true; }

	public:
		HelpOpt theHelpOpt;
		VersionOpt theVersOpt;
		StrOpt theCfgName;
		StrArrOpt theCfgDirs;
		StrOpt theZoneName;
		StrOpt theAddrStr;
		StrOpt theNameStr;
};

static String TheZoneName;
static ArraySym TheAddrs("addr");
static ArraySym TheNames("addr");
static MyOpts TheOpts;


/* MyOpt */

bool MyOpts::validate() const {
	if (theCfgName) {
		if (theZoneName)
			cerr << "--config is mutually exclusive with --zone" << endl;
		else
		if (theAddrStr || theNameStr)
			cerr << "--config is mutually exclusive with --addresses and --names" << endl;
		else
			return true;
	} else {
		if (!theZoneName)
			cerr << "must specify the zone name" << endl;
		else
		if (!theAddrStr || !theNameStr)
			cerr << "must specify addresses and domain names" << endl;
		else
			return true;
	}
	return false;
}

static
void configureLogs(int prec) {
	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);
}

static
void configureViaPgl() {
	TheOpts.theCfgDirs.copy(PglPp::TheDirs);

	PglPp pp(TheOpts.theCfgName);
	PglParser parser(&pp);

	parser.parse();

	if (PglSemx::TheAddrMapsToUse.count() != 1)
		cerr << TheOpts.theCfgName << ": must use() exactly one address map" << endl << xexit;

	const AddrMapSym *ams = PglSemx::TheAddrMapsToUse.last();
	
	TheZoneName = ams->zone();
	TheAddrs.append(*ams->addressesSym());
	TheNames.append(*ams->namesSym());

	if (!TheZoneName) {
		cerr << ams->loc() << "address map must specify zone name" 
			<< endl << xexit;
	}
	if (!TheAddrs.count()) {
		cerr << ams->loc() << "address map must specify at least one IP addresses" 
			<< endl << xexit;
	}
	if (!TheNames.count()) {
		cerr << ams->loc() << "address map must specify at least one domain name" 
			<< endl << xexit;
	}
}

static
void configureViaCmdLine() {
	PglNetAddrRange addrs;
	PglNetAddrRange names;

	if (!addrs.parse(TheOpts.theAddrStr))
		cerr << here << "malformed address range: `" <<  TheOpts.theAddrStr << "'" << endl << xexit;

	if (!names.parse(TheOpts.theNameStr))
		cerr << here << "malformed name range: `" <<  TheOpts.theNameStr << "'" << endl << xexit;

	NetAddrRangeSym as;
	as.range(new PglNetAddrRange(addrs));
	TheAddrs.add(as);

	NetAddrRangeSym ns;
	ns.range(new PglNetAddrRange(names));
	TheNames.add(ns);

	TheZoneName = TheOpts.theZoneName;
}

static
void configure() {
	configureLogs(2);
	
	if (TheOpts.theCfgName)
		configureViaPgl();
	else
		configureViaCmdLine();

	// 1:1 mapping only, for now
	if (TheAddrs.count() != TheNames.count()) {
		cerr << here << "the number of IP addresses (" << TheAddrs.count()
			<< ") differs from the number of domain names (" << TheNames.count() 
			<< ")" << endl << xexit;
	}
}

static
void printSoa(ostream &os) {
	const String serialNo = "20000833"; // XXX;

	os << "$TTL 3600" << endl << endl;

	os << "@ IN SOA "
		<< TheZoneName << '.'
		<< " root." << TheZoneName << '.' // person responsible
		<< " (" << endl
		<< "\t " << serialNo << "; Serial" << endl
		<< "\t 10800   ; Refresh" << endl
		<< "\t 1801    ; Retry" << endl
		<< "\t 3600000 ; Expire" << endl
		<< "\t 259200  ; Minimum" << endl
		<< ")" << endl
		<< endl;

	// also add NS records
	// XXX: extract addresses from Robot.dns_resolver if possible
	os << "\tIN\tNS\tns." << TheZoneName << '.' << endl;
	os << "ns\tIN\tA\t" << "172.16.101.100" << endl;

	os << endl;
}

static
String reverseAddr(const NetAddr &addr, int octets = 4) {
	char buf[1024];
	ostrstream os(buf, sizeof(buf));
	for (int i = octets; i >= 1; --i) {
		if (i != octets)
			os << '.';
		os << addr.octet(i);
	}
	os << ends;
	return buf;
}

static
String forwardName(const NetAddr &addr) {
	const String &name = addr.addrA();
	if (const char *zone = name.str(TheZoneName)) {
		const int end = zone - name - 1; // position of the dot
		if (TheZoneName == zone && end >= 0 && name[end] == '.') {
			return addr.addrA()(0, end);
		}
	}

	cerr << here << "warning: " << name << " does not end with zone suffix ("
		<< TheZoneName << "), but probably should" << endl;

	return name;
}

int main(int argc, char *argv[]) {

	CmdLine cmd;
	cmd.configure(Array<OptGrp*>() << &TheOpts);
	if (!cmd.parse(argc, argv) || !TheOpts.validate())
		return -1;

	configure();

	const String dirFname(TheZoneName);
	const String revFname(TheZoneName + ".rev");
	ofstream direct(dirFname);
	ofstream reverse(revFname);

	NetAddr commonAddr;
	const String commonName = TheZoneName;

	// SOA records
	printSoa(direct);
	printSoa(reverse);

	// A and PTR records
	for (int i = 0; i < TheAddrs.count(); ++i) {
		const NetAddrSym &as = (NetAddrSym&)TheAddrs.item(i)->cast("addr");
		const NetAddrSym &ns = (NetAddrSym&)TheNames.item(i)->cast("addr");
		const NetAddr addr(as.val().addrN(), -1); // remove port number
		const NetAddr name(ns.val().addrA(), -1); // remove port number

		if (!i)
			commonAddr = addr;

		direct 
			<< forwardName(name) << "\t\tIN\tA\t" << addr << endl;

		reverse 
			<< reverseAddr(addr) << ".IN-ADDR.ARPA."
			<< "\t\tIN\tPTR\t" << name << '.'  << endl;
	}

	cout << "# possible BIND configuration for named.conf is below " << endl
		<< "# zone files are " << revFname << " and " << dirFname << endl
		<< endl;

	cout << "options {" << endl
		<< "\tdirectory \"/etc/namedb\";" << endl
		<< "};" << endl
		<< endl;

	cout << "zone \"" << reverseAddr(commonAddr, 2) << ".IN-ADDR.ARPA\" {" << endl
		<< "\ttype master;" << endl
		<< "\tfile \"" << revFname << "\";" << endl
		<< "};" << endl
		<< endl;

	cout << "zone \"" << commonName << "\" {" << endl
		<< "\ttype master;" << endl
		<< "\tfile \"" << dirFname << "\";" << endl
		<< "};" << endl
		<< endl;

	return 0;
}
